// oVirt.js prototype.
// Requires ECMAScript 5.1 compliant environment:
//  http://www.ecma-international.org/ecma-262/5.1/
// Currently requires Lo-Dash library:
//  http://lodash.com/

// Contain the SDK within 'ovirt' global variable.
var ovirt = {};

// Expose services used by API implementation.
// Each service should be an object containing reusable methods.
// Services declared in this namespace have multiple roles:
//  1. existing services can be overridden to achieve SDK portability for given runtime environment
//  2. existing services can be used and new services can be added by SDK extensions or clients
ovirt.svc = (function () {
    'use strict';

    var services = {};

    // HTTP service for sending requests to remote server via HTTP protocol.
    // This implementation uses XMLHttpRequest object supported by web browsers.
    services.http = {
        // Send HTTP request to remote server.
        // 'httpParams' should be an object containing HTTP request parameters:
        //  - 'method': HTTP method to use, defaults to 'GET'
        //  - 'url': URL to send request to, defaults to empty string (document base URI)
        //  - 'headers': object containing extra HTTP headers, defaults to empty object
        //  - 'body': any value supported by XMLHttpRequest.send function, defaults to null
        // 'onSuccess' should be a function to invoke when HTTP request completed successfully (status code 200).
        // 'onError' should be a function to invoke when HTTP request failed (status code other than 200).
        send: function (httpParams, onSuccess, onError) {
            var xhr = new XMLHttpRequest();

            // Sanitize inputs.
            httpParams = httpParams || {};

            xhr.onreadystatechange = function () {
                if (xhr.readyState === 4) {
                    var xhrSuccess = xhr.status >= 200 && xhr.status < 300;

                    // Invoke appropriate handler function.
                    if (xhrSuccess && _.isFunction(onSuccess)) {
                        onSuccess(xhr.responseText);
                    } else if (!xhrSuccess && _.isFunction(onError)) {
                        onError(xhr.status);
                    }
                }
            };

            // Resolve HTTP request parameters.
            var method = httpParams.method || 'GET';
            var url = httpParams.url || '';
            var headers = httpParams.headers || {};
            var body = httpParams.body || null;

            // Initiate new request.
            xhr.open(method, url, true);

            // Include credentials such as HTTP auth information and cookies in cross-site requests.
            xhr.withCredentials = true;

            // Apply request headers.
            Object.keys(headers).forEach(function (name) {
                var value = headers[name] && headers[name].toString();
                xhr.setRequestHeader(name, value || '');
            });

            // Fire off the request.
            xhr.send(body);
        }
    };

    return services;
})();

// Expose utilities used by API implementation.
// Unlike services, utilities are meant to assist with writing more efficient JavaScript code.
ovirt.util = (function (services) {
    'use strict';

    var utilities = {};

    // Utility for creating objects using mixins with dependency resolution mechanism.
    utilities.mixins = {
        // Construct an object using mixin functions and dependencies.
        // 'spec' should be an object containing information needed for object creation.
        // 'mixins' should be an array object containing 0 or more mixin functions.
        // 'deps' should be an array object containing 0 or more string values identifying required dependencies.
        // This function performs following steps:
        //  1. create 'my' object, representing shared secrets among mixins
        //  2. create 'that' object, representing actual return value
        //  3. augment 'my' object with dependencies resolved from service namespace
        //  4. apply mixins on 'that' object, binding 'this' to 'that' object
        //  5. return 'that' object
        mix: function (spec, mixins, deps) {
            var my = {};
            var that = {};

            // Sanitize inputs.
            spec = spec || {};
            mixins = mixins || [];
            deps = deps || [];

            // Resolve dependencies and augment 'my'.
            deps.forEach(function (name) {
                my[name] = services[name];
            });

            // Apply mixins on 'that', binding 'this' to 'that'.
            mixins.forEach(function (fn) {
                fn.call(that, that, spec, my);
            });

            return that;
        },

        // Mixin for objects having arbitrary options.
        withOptions: function (that, spec) {
            var options = spec.options || {};

            // Access options via getter/setter function with merge-on-set behavior.
            that.options = utilities.fns.makeGetSetMergeFn(that, options);

            return that;
        }
    };

    // Utility for working with functions.
    utilities.fns = {
        // Make a function acting as getter and setter on object value with merge-on-set behavior.
        // Calling resulting function with an object argument will merge properties of that object into 'target' object and return 'that'.
        // Otherwise, 'target' object will be returned.
        makeGetSetMergeFn: function (that, target) {
            return function (update) {
                if (_.isObject(update)) {
                    _.merge(target, update);
                    return that;
                } else {
                    return target;
                }
            };
        }
    };

    // Utility for working with objects.
    utilities.objects = {
        // Make a deep clone of 'source', merge properties of 'update' into that clone and return the clone.
        deepCloneAndMerge: function (source, update) {
            var clone = _.cloneDeep(source);
            return _.merge(clone, update);
        }
    };

    return utilities;
})(ovirt.svc);

// Expose API as the main entry point to SDK functionality.
ovirt.api = (function (utilities) {
    'use strict';

    var mixins = utilities.mixins;
    var objects = utilities.objects;
    var fns = utilities.fns;

    // Mixin for objects representing a RESTful resource collection.
    function asResourceCollection (that, spec) {
        that.get = function (id, options) {
            return makeOperation(objects.deepCloneAndMerge(spec, {
                options: options,
                httpMethod: 'GET',
                apiContextPath: spec.apiContextPath + '/' + id,
                transformResult: function (data) {
                    return makeResource({
                        data: transformResourceData(data)
                    });
                }
            }));
        };

        // Return operation to list resources grouped by this resource collection.
        that.list = function (options) {
            return makeOperation(objects.deepCloneAndMerge(spec, {
                options: options,
                httpMethod: 'GET',
                transformResult: makeResourceArray
            }));
        };

        // Return operation to add resource represented by 'data' object to this resource collection.
        that.add = function (data, options) {
            return makeOperation(objects.deepCloneAndMerge(spec, {
                options: options,
                httpMethod: 'POST',
                httpBody: JSON.stringify(data),
                transformResult: function (data) {
                    return makeResource({
                        data: transformResourceData(data)
                    });
                }
            }));
        };
    }

    // Transform raw resource data for use by resource objects.
    function transformResourceData (data) {
        function definePropertyReadOnly (propName, propValue) {
            Object.defineProperty(data, propName, {
                value: propValue,
                writable: false,
                configurable: true,
                enumerable: true
            });
        }

        function reduceLinks (array) {
            return array.reduce(function (result, link) {
                result[link.rel] = link.href;
                return result;
            }, {});
        }

        // Make 'id' non-writable.
        definePropertyReadOnly('id', data.id);

        // Make 'href' non-writable.
        definePropertyReadOnly('href', data.href);

        // Transform and expose nested resource collections.
        var subCollectionLinks = data.link || [];
        definePropertyReadOnly('subCollections', reduceLinks(subCollectionLinks));
        delete data.link;

        // Transform and expose supported actions.
        var actionLinks = (data.actions && data.actions.link) || [];
        definePropertyReadOnly('actions', reduceLinks(actionLinks));

        return data;
    }

    // Extract raw resource data from 'obj' and return array of resource objects.
    function makeResourceArray (obj) {
        // Extract resource data, 'obj' should contain a single enumerable property.
        var keys = Object.keys(obj);
        var dataArray = (keys.length === 1) ? obj[keys[0]] : [];

        // Sanitize resource data.
        dataArray = Array.isArray(dataArray) ? dataArray : [];

        // Return resource objects.
        return dataArray.map(function (data) {
            return makeResource({
                data: transformResourceData(data)
            });
        });
    }

    function makeResourceCollection (spec) {
        return mixins.mix(spec, [mixins.withOptions, asResourceCollection]);
    }

    // Compute change between two objects and return a diff object containing the difference.
    // The diff object will be null if there is no change.
    function computeDiff (original, modified) {
        var diff = {};

        if (_.isObject(original) && _.isObject(modified)) {
            Object.keys(modified).forEach(function (key) {
                var propertyDiff;

                if (!original.hasOwnProperty(key)) {
                    // Add new property missing in 'original' automatically into diff object.
                    diff[key] = modified[key];
                } else {
                    // Otherwise, compute change for existing property recursively and update diff object.
                    propertyDiff = computeDiff(original[key], modified[key]);
                    if (propertyDiff) {
                        diff[key] = propertyDiff;
                    }
                }
            });
            return _.isEqual(diff, {}) ? null : diff;
        } else {
            return _.isEqual(original, modified) ? null : modified;
        }
    }

    // Mixin for objects representing a RESTful resource.
    function asResource (that, spec) {
        // Raw resource data stored in object representation.
        var data, originalData;

        function assignData (resourceData) {
            data = resourceData || {};
            originalData = _.cloneDeep(data);

            // Access resource data via getter/setter function with merge-on-set behavior.
            // TODO prevent setting new properties (ones not already present in full data representation)
            that.data = fns.makeGetSetMergeFn(that, data);
        }

        // Initialize resource data.
        assignData(spec.data);

        // Return operation to update this resource.
        that.update = function (options) {
            return makeOperation(objects.deepCloneAndMerge(spec, {
                options: options,
                httpMethod: 'PUT',
                httpBody: JSON.stringify(computeDiff(originalData, data)),
                transformResult: function (data) {
                    assignData(transformResourceData(data));
                    return that;
                }
            }));
        };

        // Return operation to delete this resource.
        that.delete = function (options) {
            return makeOperation(objects.deepCloneAndMerge(spec, {
                options: options,
                httpMethod: 'DELETE',
                httpBody: '{}'
            }));
        };

        // Augment resource with nested resource collections.
        var subCollections = data.subCollections || {};
        Object.keys(subCollections).forEach(function (name) {
            that[name] = makeResourceCollection({
                apiContextPath: subCollections[name]
            });
        });

        // Augment resource with supported actions.
        var actions = data.actions || {};
        Object.keys(actions).forEach(function (name) {
            // Return operation to invoke given action using 'actionData' on this resource.
            that[name] = function (actionData, options) {
                return makeOperation(objects.deepCloneAndMerge(spec, {
                    options: options,
                    httpMethod: 'POST',
                    httpBody: JSON.stringify(actionData),
                    apiContextPath: actions[name]
                }));
            };
        });

        // TODO expose top-level resource data properties for convenience
    }

    function makeResource (spec) {
        return mixins.mix(_.merge(spec, {
            apiContextPath: spec.data && spec.data.href
        }), [mixins.withOptions, asResource]);
    }

    // If 'fn' is a function, return that function.
    // Otherwise, return a no-op function.
    function getFnOrNoop (fn) {
        return _.isFunction(fn) ? fn : _.noop;
    }

    // Mixin for objects implementing a RESTful operation.
    // Requires 'http' service.
    function asOperation (that, spec, my) {
        var onSuccess = _.noop;
        var onError = _.noop;

        function httpSuccess (responseText) {
            try {
                // Parse response as JSON.
                var result = JSON.parse(responseText);

                // Transform resulting object, if necessary.
                if (_.isFunction(spec.transformResult)) {
                    result = spec.transformResult(result);
                }

                // Invoke operation success callback.
                onSuccess(result);
            } catch (e) {
                // TODO handle error properly
                console.log('Error while parsing response: ' + e);
            }
        }

        function httpError (statusCode) {
            // TODO handle error properly
            console.log('Request failed with status code ' + statusCode);
        }

        // Register operation success callback.
        that.success = function (callback) {
            onSuccess = getFnOrNoop(callback);
            return that;
        };

        // Register operation error callback.
        that.error = function (callback) {
            onError = getFnOrNoop(callback);
            return that;
        };

        // Run the operation, sending HTTP request to remote server.
        that.run = function () {
            // Prepare request headers.
            var headers = spec.httpHeaders || {};
            headers['Accept'] = 'application/json';
            headers['Content-Type'] = 'application/json';
            headers['Filter'] = api.options().filterResults;

            // Prepare request parameters.
            var httpParams = {
                method: spec.httpMethod,
                url: api.options().engineBaseUrl + spec.apiContextPath,
                headers: headers,
                body: spec.httpBody
            };

            // Send the request.
            my.http.send(httpParams, httpSuccess, httpError);
        };
    }

    function makeOperation (spec) {
        return mixins.mix(spec, [mixins.withOptions, asOperation], ['http']);
    }

    var api = mixins.mix({
        options: {
            // Engine base URL, without a trailing slash.
            engineBaseUrl: 'http://127.0.0.1:8080',
            // Flag indicating whether to filter results based on user's permissions.
            filterResults: false
        }
    }, [mixins.withOptions]);

    // Augment API with top-level resource collections.
    // TODO implement api.init function that discovers all top-level collections
    api.datacenters = makeResourceCollection({
        apiContextPath: '/ovirt-engine/api/datacenters'
    });

    return api;
})(ovirt.util);
